@@ -1,4 +1,6 @@
-"""Parses and creates Grammar objects"""
+"""Parses and compiles Lark grammars into an internal representation.
+"""
+
 import hashlib
 import os.path
 import sys
@@ -7,10 +9,10 @@ from copy import copy, deepcopy
 import pkgutil
 from ast import literal_eval
 from contextlib import suppress
-from typing import List, Tuple, Union, Callable, Dict, Optional, Sequence
+from typing import List, Tuple, Union, Callable, Dict, Optional, Sequence, Generator
 
-from .utils import bfs, logger, classify_bool, is_id_continue, is_id_start, bfs_all_unique, small_factors
-from .lexer import Token, TerminalDef, PatternStr, PatternRE
+from .utils import bfs, logger, classify_bool, is_id_continue, is_id_start, bfs_all_unique, small_factors, OrderedSet
+from .lexer import Token, TerminalDef, PatternStr, PatternRE, Pattern
 
 from .parse_tree_builder import ParseTreeBuilder
 from .parser_frontends import ParsingFrontend
@@ -23,7 +25,6 @@ from .tree import Tree, SlottedTree as ST
 from .visitors import Transformer, Visitor, v_args, Transformer_InPlace, Transformer_NonRecursive
 inline_args = v_args(inline=True)
 
-__path__ = os.path.dirname(__file__)
 IMPORT_PATHS = ['grammars']
 
 EXT = '.lark'
@@ -79,7 +80,7 @@ TERMINALS = {
     '_RBRA': r'\]',
     '_LBRACE': r'\{',
     '_RBRACE': r'\}',
-    'OP': '[+*]|[?](?![a-z])',
+    'OP': '[+*]|[?](?![a-z_])',
     '_COLON': ':',
     '_COMMA': ',',
     '_OR': r'\|',
@@ -94,7 +95,7 @@ TERMINALS = {
     '_NL': r'(\r?\n)+\s*',
     '_NL_OR': r'(\r?\n)+\s*\|',
     'WS': r'[ \t]+',
-    'COMMENT': r'\s*//[^\n]*',
+    'COMMENT': r'\s*//[^\n]*|\s*#[^\n]*',
     'BACKSLASH': r'\\[ ]*\n',
     '_TO': '->',
     '_IGNORE': r'%ignore',
@@ -194,10 +195,10 @@ REPEAT_BREAK_THRESHOLD = 50
 
 
 class FindRuleSize(Transformer):
-    def __init__(self, keep_all_tokens):
+    def __init__(self, keep_all_tokens: bool):
         self.keep_all_tokens = keep_all_tokens
 
-    def _will_not_get_removed(self, sym):
+    def _will_not_get_removed(self, sym: Symbol) -> bool:
         if isinstance(sym, NonTerminal):
             return not sym.name.startswith('_')
         if isinstance(sym, Terminal):
@@ -206,7 +207,7 @@ class FindRuleSize(Transformer):
             return False
         assert False, sym
 
-    def _args_as_int(self, args):
+    def _args_as_int(self, args: List[Union[int, Symbol]]) -> Generator[int, None, None]:
         for a in args:
             if isinstance(a, int):
                 yield a
@@ -215,10 +216,10 @@ class FindRuleSize(Transformer):
             else:
                 assert False
 
-    def expansion(self, args):
+    def expansion(self, args) -> int:
         return sum(self._args_as_int(args))
 
-    def expansions(self, args):
+    def expansions(self, args) -> int:
         return max(self._args_as_int(args))
 
 
@@ -231,7 +232,7 @@ class EBNF_to_BNF(Transformer_InPlace):
         self.i = 0
         self.rule_options = None
 
-    def _name_rule(self, inner):
+    def _name_rule(self, inner: str):
         new_name = '__%s_%s_%d' % (self.prefix, inner, self.i)
         self.i += 1
         return new_name
@@ -242,7 +243,7 @@ class EBNF_to_BNF(Transformer_InPlace):
         self.rules_cache[key] = t
         return t
 
-    def _add_recurse_rule(self, type_, expr):
+    def _add_recurse_rule(self, type_: str, expr: Tree):
         try:
             return self.rules_cache[expr]
         except KeyError:
@@ -311,14 +312,14 @@ class EBNF_to_BNF(Transformer_InPlace):
             ])
             return self._add_rule(key, new_name, tree)
 
-    def _generate_repeats(self, rule, mn, mx):
+    def _generate_repeats(self, rule: Tree, mn: int, mx: int):
         """Generates a rule tree that repeats ``rule`` exactly between ``mn`` to ``mx`` times.
         """
         # For a small number of repeats, we can take the naive approach
         if mx < REPEAT_BREAK_THRESHOLD:
             return ST('expansions', [ST('expansion', [rule] * n) for n in range(mn, mx + 1)])
 
-        # For large repeat values, we break the repetition into sub-rules. 
+        # For large repeat values, we break the repetition into sub-rules.
         # We treat ``rule~mn..mx`` as ``rule~mn rule~0..(diff=mx-mn)``.
         # We then use small_factors to split up mn and diff up into values [(a, b), ...]
         # This values are used with the help of _add_repeat_rule and _add_repeat_rule_opt
@@ -342,7 +343,7 @@ class EBNF_to_BNF(Transformer_InPlace):
 
         return ST('expansions', [ST('expansion', [mn_target] + [diff_opt_target])])
 
-    def expr(self, rule, op, *args):
+    def expr(self, rule: Tree, op: Token, *args):
         if op.value == '?':
             empty = ST('expansion', [])
             return ST('expansions', [rule, empty])
@@ -371,7 +372,7 @@ class EBNF_to_BNF(Transformer_InPlace):
 
         assert False, op
 
-    def maybe(self, rule):
+    def maybe(self, rule: Tree):
         keep_all_tokens = self.rule_options and self.rule_options.keep_all_tokens
         rule_size = FindRuleSize(keep_all_tokens).transform(rule)
         empty = ST('expansion', [_EMPTY] * rule_size)
@@ -381,11 +382,11 @@ class EBNF_to_BNF(Transformer_InPlace):
 class SimplifyRule_Visitor(Visitor):
 
     @staticmethod
-    def _flatten(tree):
+    def _flatten(tree: Tree):
         while tree.expand_kids_by_data(tree.data):
             pass
 
-    def expansion(self, tree):
+    def expansion(self, tree: Tree):
         # rules_list unpacking
         # a : b (c|d) e
         #  -->
@@ -416,7 +417,7 @@ class SimplifyRule_Visitor(Visitor):
             tree.data = 'expansions'
             tree.children = aliases
 
-    def expansions(self, tree):
+    def expansions(self, tree: Tree):
         self._flatten(tree)
         # Ensure all children are unique
         if len(set(tree.children)) != len(tree.children):
@@ -609,7 +610,7 @@ class PrepareLiterals(Transformer_InPlace):
         return ST('pattern', [PatternRE(regexp)])
 
 
-def _make_joined_pattern(regexp, flags_set):
+def _make_joined_pattern(regexp, flags_set) -> PatternRE:
     return PatternRE(regexp, ())
 
 class TerminalTreeToPattern(Transformer_NonRecursive):
@@ -617,15 +618,17 @@ class TerminalTreeToPattern(Transformer_NonRecursive):
         p ,= ps
         return p
 
-    def expansion(self, items):
-        assert items
+    def expansion(self, items: List[Pattern]) -> Pattern:
+        if not items:
+            return PatternStr('')
+
         if len(items) == 1:
             return items[0]
 
         pattern = ''.join(i.to_regexp() for i in items)
         return _make_joined_pattern(pattern, {i.flags for i in items})
 
-    def expansions(self, exps):
+    def expansions(self, exps: List[Pattern]) -> Pattern:
         if len(exps) == 1:
             return exps[0]
 
@@ -636,7 +639,8 @@ class TerminalTreeToPattern(Transformer_NonRecursive):
         pattern = '(?:%s)' % ('|'.join(i.to_regexp() for i in exps))
         return _make_joined_pattern(pattern, {i.flags for i in exps})
 
-    def expr(self, args):
+    def expr(self, args) -> Pattern:
+        inner: Pattern
         inner, op = args[:2]
         if op == '~':
             if len(args) == 3:
@@ -683,7 +687,7 @@ class Grammar:
         self.rule_defs = rule_defs
         self.ignore = ignore
 
-    def compile(self, start, terminals_to_keep):
+    def compile(self, start, terminals_to_keep) -> Tuple[List[TerminalDef], List[Rule], List[str]]:
         # We change the trees in-place (to support huge grammars)
         # So deepcopy allows calling compile more than once.
         term_defs = [(n, (nr_deepcopy_tree(t), p)) for n, (t, p) in self.term_defs]
@@ -732,7 +736,7 @@ class Grammar:
             ebnf_to_bnf.prefix = name
             anon_tokens_transf.rule_options = rule_options
             tree = transformer.transform(rule_tree)
-            res = ebnf_to_bnf.transform(tree)
+            res: Tree = ebnf_to_bnf.transform(tree)
             rules.append((name, res, options))
         rules += ebnf_to_bnf.new_rules
 
@@ -742,7 +746,7 @@ class Grammar:
         rule_tree_to_text = RuleTreeToText()
 
         simplify_rule = SimplifyRule_Visitor()
-        compiled_rules = []
+        compiled_rules: List[Rule] = []
         for rule_content in rules:
             name, tree, options = rule_content
             simplify_rule.visit(tree)
@@ -752,7 +756,7 @@ class Grammar:
                 if alias and name.startswith('_'):
                     raise GrammarError("Rule %s is marked for expansion (it starts with an underscore) and isn't allowed to have aliases (alias=%s)"% (name, alias))
 
-                empty_indices = [x==_EMPTY for x in expansion]
+                empty_indices = tuple(x==_EMPTY for x in expansion)
                 if any(empty_indices):
                     exp_options = copy(options) or RuleOptions()
                     exp_options.empty_indices = empty_indices
@@ -763,6 +767,7 @@ class Grammar:
                 for sym in expansion:
                     assert isinstance(sym, Symbol)
                     if sym.is_term and exp_options and exp_options.keep_all_tokens:
+                        assert isinstance(sym, Terminal)
                         sym.filter_out = False
                 rule = Rule(NonTerminal(name), expansion, i, alias, exp_options)
                 compiled_rules.append(rule)
@@ -780,7 +785,7 @@ class Grammar:
                     assert len({(r.alias, r.order, r.options) for r in dups}) == len(dups)
 
             # Remove duplicates
-            compiled_rules = list(set(compiled_rules))
+            compiled_rules = list(OrderedSet(compiled_rules))
 
         # Filter out unused rules
         while True:
@@ -1050,6 +1055,8 @@ def _make_rule_tuple(modifiers_tree, name, params, priority_tree, expansions):
     if modifiers_tree.children:
         m ,= modifiers_tree.children
         expand1 = '?' in m
+        if expand1 and name.startswith('_'):
+            raise GrammarError("Inlined rules (_rule) cannot use the ?rule modifier.")
         keep_all_tokens = '!' in m
     else:
         keep_all_tokens = False
@@ -1233,7 +1240,7 @@ class GrammarBuilder:
         tree = _parse_grammar(grammar_text, grammar_name)
 
         imports: Dict[Tuple[str, ...], Tuple[Optional[str], Dict[str, str]]] = {}
-          
+
         for stmt in tree.children:
             if stmt.data == 'import':
                 dotted_path, base_path, aliases = self._unpack_import(stmt, grammar_name)
@@ -1312,11 +1319,11 @@ class GrammarBuilder:
             except IOError:
                 continue
             else:
-                h = hashlib.md5(text.encode('utf8')).hexdigest()
+                h = sha256_digest(text)
                 if self.used_files.get(joined_path, h) != h:
                     raise RuntimeError("Grammar file was changed during importing")
                 self.used_files[joined_path] = h
-                    
+
                 gb = GrammarBuilder(self.global_keep_all_tokens, self.import_paths, self.used_files)
                 gb.load_grammar(text, joined_path, mangle)
                 gb._remove_unused(map(mangle, aliases))
@@ -1390,8 +1397,8 @@ def verify_used_files(file_hashes):
                 text = pkgutil.get_data(*path).decode('utf-8')
         if text is None: # We don't know how to load the path. ignore it.
             continue
-            
-        current = hashlib.md5(text.encode()).hexdigest()
+
+        current = sha256_digest(text)
         if old != current:
             logger.info("File %r changed, rebuilding Parser" % path)
             return False
@@ -1407,3 +1414,15 @@ def load_grammar(grammar, source, import_paths, global_keep_all_tokens):
     builder = GrammarBuilder(global_keep_all_tokens, import_paths)
     builder.load_grammar(grammar, source)
     return builder.build(), builder.used_files
+
+
+def sha256_digest(s: str) -> str:
+    """Get the sha256 digest of a string
+
+    Supports the `usedforsecurity` argument for Python 3.9+ to allow running on
+    a FIPS-enabled system.
+    """
+    if sys.version_info >= (3, 9):
+        return hashlib.sha256(s.encode('utf8'), usedforsecurity=False).hexdigest()
+    else:
+        return hashlib.sha256(s.encode('utf8')).hexdigest()
