@@ -1,25 +1,29 @@
 from __future__ import annotations
 
-import copy
-import re
+import warnings
 
-from contextlib import contextmanager
-from pathlib import Path
 from typing import TYPE_CHECKING
-from typing import Collection
-from typing import Iterable
-from typing import Iterator
+from typing import ClassVar
 from typing import TypeVar
-from typing import cast
+
+from packaging.utils import canonicalize_name
 
 from poetry.core.constraints.version import parse_constraint
+from poetry.core.constraints.version.exceptions import ParseConstraintError
 from poetry.core.packages.dependency_group import MAIN_GROUP
 from poetry.core.packages.specification import PackageSpecification
-from poetry.core.packages.utils.utils import create_nested_marker
-from poetry.core.version.markers import parse_marker
+from poetry.core.utils.patterns import AUTHOR_REGEX
+from poetry.core.version.exceptions import InvalidVersionError
 
 
 if TYPE_CHECKING:
+    from collections.abc import Collection
+    from collections.abc import Iterable
+    from collections.abc import Mapping
+    from collections.abc import Sequence
+    from pathlib import Path
+
+    from packaging.licenses import NormalizedLicenseExpression
     from packaging.utils import NormalizedName
 
     from poetry.core.constraints.version import Version
@@ -31,11 +35,9 @@ if TYPE_CHECKING:
 
     T = TypeVar("T", bound="Package")
 
-AUTHOR_REGEX = re.compile(r"(?u)^(?P<name>[- .,\w\d'’\"():&]+)(?: <(?P<email>.+?)>)?$")
-
 
 class Package(PackageSpecification):
-    AVAILABLE_PYTHONS = {
+    AVAILABLE_PYTHONS: ClassVar[set[str]] = {
         "2",
         "2.7",
         "3",
@@ -47,13 +49,15 @@ class Package(PackageSpecification):
         "3.9",
         "3.10",
         "3.11",
+        "3.12",
+        "3.13",
+        "3.14",
     }
 
     def __init__(
         self,
         name: str,
         version: str | Version,
-        pretty_version: str | None = None,
         source_type: str | None = None,
         source_url: str | None = None,
         source_reference: str | None = None,
@@ -78,36 +82,44 @@ class Package(PackageSpecification):
             features=features,
         )
 
-        self._set_version(version, pretty_version)
+        # Attributes must be immutable for clone() to be safe!
+        # (For performance reasons, clone only creates a copy instead of a deep copy).
+
+        self._set_version(version)
 
         self.description = ""
 
-        self._authors: list[str] = []
-        self._maintainers: list[str] = []
+        self.authors: Sequence[str] = []
+        self.maintainers: Sequence[str] = []
 
         self.homepage: str | None = None
         self.repository_url: str | None = None
         self.documentation_url: str | None = None
-        self.keywords: list[str] = []
+        self.keywords: Sequence[str] = []
         self._license: License | None = None
+        self._license_expression: NormalizedLicenseExpression | None = None
+        # meaning of different values:
+        # - tuple: project.license-files -> NO default handling
+        #   - empty tuple: explicitly no files!
+        # - None: nothing specified -> default handling
+        # - Path: deprecated project.license.file -> file + default handling
+        self.license_files: tuple[str, ...] | Path | None = None
         self.readmes: tuple[Path, ...] = ()
+        self.readme_content_type: str | None = None
+        self.readme_content: str | None = None
 
-        self.extras: dict[NormalizedName, list[Dependency]] = {}
+        self.extras: Mapping[NormalizedName, Sequence[Dependency]] = {}
 
-        self._dependency_groups: dict[str, DependencyGroup] = {}
+        self._dependency_groups: Mapping[NormalizedName, DependencyGroup] = {}
 
-        # For compatibility with previous version, we keep the category
-        self.category = "main"
-        self.files: list[dict[str, str]] = []
+        self.files: Sequence[Mapping[str, str]] = []
         self.optional = False
 
-        self.classifiers: list[str] = []
+        self.classifiers: Sequence[str] = []
 
         self._python_versions = "*"
         self._python_constraint = parse_constraint("*")
-        self._python_marker: BaseMarker = AnyMarker()
 
-        self.platform = None
         self.marker: BaseMarker = AnyMarker()
 
         self.root_dir: Path | None = None
@@ -130,7 +142,7 @@ class Package(PackageSpecification):
 
     @property
     def pretty_version(self) -> str:
-        return self._pretty_version
+        return self._version.text
 
     @property
     def unique_name(self) -> str:
@@ -145,27 +157,23 @@ class Package(PackageSpecification):
 
     @property
     def full_pretty_version(self) -> str:
-        if self.source_type in ["file", "directory", "url"]:
-            return f"{self._pretty_version} {self.source_url}"
+        if self.source_type in ("file", "directory", "url"):
+            return f"{self.pretty_version} {self.source_url}"
 
-        if self.source_type not in ["hg", "git"]:
-            return self._pretty_version
+        if self.source_type not in ("hg", "git"):
+            return self.pretty_version
 
         ref: str | None
         if self.source_resolved_reference and len(self.source_resolved_reference) == 40:
             ref = self.source_resolved_reference[0:7]
-            return f"{self._pretty_version} {ref}"
+            return f"{self.pretty_version} {ref}"
 
         # if source reference is a sha1 hash -- truncate
         if self.source_reference and len(self.source_reference) == 40:
-            return f"{self._pretty_version} {self.source_reference[0:7]}"
+            return f"{self.pretty_version} {self.source_reference[0:7]}"
 
         ref = self._source_resolved_reference or self._source_reference
-        return f"{self._pretty_version} {ref}"
-
-    @property
-    def authors(self) -> list[str]:
-        return self._authors
+        return f"{self.pretty_version} {ref}"
 
     @property
     def author_name(self) -> str | None:
@@ -175,10 +183,6 @@ class Package(PackageSpecification):
     def author_email(self) -> str | None:
         return self._get_author()["email"]
 
-    @property
-    def maintainers(self) -> list[str]:
-        return self._maintainers
-
     @property
     def maintainer_name(self) -> str | None:
         return self._get_maintainer()["name"]
@@ -190,7 +194,7 @@ class Package(PackageSpecification):
     @property
     def requires(self) -> list[Dependency]:
         """
-        Returns the main dependencies
+        Returns the main dependencies.
         """
         if not self._dependency_groups or MAIN_GROUP not in self._dependency_groups:
             return []
@@ -198,35 +202,35 @@ class Package(PackageSpecification):
         return self._dependency_groups[MAIN_GROUP].dependencies
 
     @property
-    def all_requires(
-        self,
-    ) -> list[Dependency]:
+    def all_requires(self) -> list[Dependency]:
         """
-        Returns the main dependencies and group dependencies.
+        Returns the main dependencies and group dependencies
+        enriched with Poetry-specific information for locking.
         """
         return [
             dependency
             for group in self._dependency_groups.values()
-            for dependency in group.dependencies
+            for dependency in group.dependencies_for_locking
         ]
 
-    def _set_version(
-        self, version: str | Version, pretty_version: str | None = None
-    ) -> None:
+    def _set_version(self, version: str | Version) -> None:
         from poetry.core.constraints.version import Version
 
         if not isinstance(version, Version):
-            self._version = Version.parse(version)
-            self._pretty_version = pretty_version or version
-        else:
-            self._version = version
-            self._pretty_version = pretty_version or self._version.text
+            try:
+                version = Version.parse(version)
+            except InvalidVersionError:
+                raise InvalidVersionError(
+                    f"Invalid version '{version}' on package {self.name}"
+                )
+
+        self._version = version
 
     def _get_author(self) -> dict[str, str | None]:
-        if not self._authors:
+        if not self.authors:
             return {"name": None, "email": None}
 
-        m = AUTHOR_REGEX.match(self._authors[0])
+        m = AUTHOR_REGEX.match(self.authors[0])
 
         if m is None:
             raise ValueError(
@@ -240,10 +244,10 @@ class Package(PackageSpecification):
         return {"name": name, "email": email}
 
     def _get_maintainer(self) -> dict[str, str | None]:
-        if not self._maintainers:
+        if not self.maintainers:
             return {"name": None, "email": None}
 
-        m = AUTHOR_REGEX.match(self._maintainers[0])
+        m = AUTHOR_REGEX.match(self.maintainers[0])
 
         if m is None:
             raise ValueError(
@@ -262,11 +266,16 @@ class Package(PackageSpecification):
 
     @python_versions.setter
     def python_versions(self, value: str) -> None:
+        try:
+            constraint = parse_constraint(value)
+        except ParseConstraintError:
+            raise ParseConstraintError(f"Invalid python versions '{value}' on {self}")
+
+        if constraint.is_empty():
+            raise ParseConstraintError(f"Python versions '{value}' on {self} is empty")
+
         self._python_versions = value
-        self._python_constraint = parse_constraint(value)
-        self._python_marker = parse_marker(
-            create_nested_marker("python_version", self._python_constraint)
-        )
+        self._python_constraint = constraint
 
     @property
     def python_constraint(self) -> VersionConstraint:
@@ -274,7 +283,18 @@ class Package(PackageSpecification):
 
     @property
     def python_marker(self) -> BaseMarker:
-        return self._python_marker
+        from poetry.core.packages.utils.utils import create_nested_marker
+        from poetry.core.version.markers import parse_marker
+
+        warnings.warn(
+            "`python_marker` is deprecated and will be removed in a future release.",
+            DeprecationWarning,
+            stacklevel=2,
+        )
+
+        return parse_marker(
+            create_nested_marker("python_version", self._python_constraint)
+        )
 
     @property
     def license(self) -> License | None:
@@ -285,16 +305,33 @@ class Package(PackageSpecification):
         from poetry.core.spdx.helpers import license_by_id
         from poetry.core.spdx.license import License
 
+        if value is not None and self.license_expression is not None:
+            raise ValueError(
+                "Cannot set license when license_expression is already set."
+            )
+
         if value is None or isinstance(value, License):
             self._license = value
         else:
             self._license = license_by_id(value)
 
+    @property
+    def license_expression(self) -> NormalizedLicenseExpression | None:
+        return self._license_expression
+
+    @license_expression.setter
+    def license_expression(self, value: NormalizedLicenseExpression | None) -> None:
+        if value is not None and self._license is not None:
+            raise ValueError(
+                "Cannot set license_expression when license is already set."
+            )
+        self._license_expression = value
+
     @property
     def all_classifiers(self) -> list[str]:
         from poetry.core.constraints.version import Version
 
-        classifiers = copy.copy(self.classifiers)
+        classifiers = list(self.classifiers)
 
         # Automatically set python classifiers
         if self.python_versions == "*":
@@ -322,13 +359,16 @@ class Package(PackageSpecification):
 
         # Automatically set license classifiers
         if self.license:
+            # License classifiers have been deprecated in PEP 639.
+            # We only use them for licenses from the deprecated [project.license] table
+            # (via self.license) and not if self.license_expression is set.
             classifiers.append(self.license.classifier)
 
         # Sort classifiers and insert python classifiers at the right location. We do
         # it like this so that 3.10 is sorted after 3.9.
         sorted_classifiers = []
         python_classifiers_inserted = False
-        for classifier in sorted(set(classifiers)):
+        for classifier in sorted(set(classifiers) - set(python_classifiers)):
             if (
                 not python_classifiers_inserted
                 and classifier > python_classifier_prefix
@@ -357,28 +397,6 @@ class Package(PackageSpecification):
 
         return urls
 
-    @property
-    def readme(self) -> Path | None:
-        import warnings
-
-        warnings.warn(
-            "`readme` is deprecated: you are getting only the first readme file. Please"
-            " use the plural form `readmes`.",
-            DeprecationWarning,
-        )
-        return next(iter(self.readmes), None)
-
-    @readme.setter
-    def readme(self, path: Path) -> None:
-        import warnings
-
-        warnings.warn(
-            "`readme` is deprecated. Please assign a tuple to the plural form"
-            " `readmes`.",
-            DeprecationWarning,
-        )
-        self.readmes = (path,)
-
     @property
     def yanked(self) -> bool:
         return isinstance(self._yanked, str) or bool(self._yanked)
@@ -395,7 +413,9 @@ class Package(PackageSpecification):
     def is_root(self) -> bool:
         return False
 
-    def dependency_group_names(self, include_optional: bool = False) -> set[str]:
+    def dependency_group_names(
+        self, include_optional: bool = False
+    ) -> set[NormalizedName]:
         return {
             name
             for name, group in self._dependency_groups.items()
@@ -403,16 +423,18 @@ class Package(PackageSpecification):
         }
 
     def add_dependency_group(self, group: DependencyGroup) -> None:
-        self._dependency_groups[group.name] = group
+        groups = dict(self._dependency_groups)
+        groups[group.name] = group
+        self._dependency_groups = groups
 
     def has_dependency_group(self, name: str) -> bool:
-        return name in self._dependency_groups
+        return canonicalize_name(name) in self._dependency_groups
 
     def dependency_group(self, name: str) -> DependencyGroup:
         if not self.has_dependency_group(name):
             raise ValueError(f'The dependency group "{name}" does not exist.')
 
-        return self._dependency_groups[name]
+        return self._dependency_groups[canonicalize_name(name)]
 
     def add_dependency(
         self,
@@ -433,11 +455,15 @@ class Package(PackageSpecification):
         """
         Returns a clone of the package with the given dependency groups excluded.
         """
-        package = self.clone()
+        canonicalized_groups = {canonicalize_name(group) for group in groups}
+        updated_groups = {
+            group_name: group
+            for group_name, group in self._dependency_groups.items()
+            if group_name not in canonicalized_groups
+        }
 
-        for group_name in groups:
-            if group_name in package._dependency_groups:
-                del package._dependency_groups[group_name]
+        package = self.clone()
+        package._dependency_groups = updated_groups
 
         return package
 
@@ -445,11 +471,13 @@ class Package(PackageSpecification):
         """
         Returns a clone of the package without optional dependency groups.
         """
+        updated_groups = {
+            group_name: group
+            for group_name, group in self._dependency_groups.items()
+            if not group.is_optional()
+        }
         package = self.clone()
-
-        for group_name, group in self._dependency_groups.items():
-            if group.is_optional():
-                del package._dependency_groups[group_name]
+        package._dependency_groups = updated_groups
 
         return package
 
@@ -464,11 +492,15 @@ class Package(PackageSpecification):
 
         If `only` is set to True, then only the given groups will be selected.
         """
+        canonicalized_groups = {canonicalize_name(group) for group in groups}
+        updated_groups = {
+            group_name: group
+            for group_name, group in self._dependency_groups.items()
+            if group_name in canonicalized_groups
+            or (not only and not group.is_optional())
+        }
         package = self.clone()
-
-        for group_name, group in self._dependency_groups.items():
-            if (only or group.is_optional()) and group_name not in groups:
-                del package._dependency_groups[group_name]
+        package._dependency_groups = updated_groups
 
         return package
 
@@ -483,9 +515,10 @@ class Package(PackageSpecification):
 
         dep: Dependency
         if self.source_type == "directory":
+            assert self._source_url is not None
             dep = DirectoryDependency(
                 self._name,
-                Path(cast(str, self._source_url)),
+                Path(self._source_url),
                 groups=list(self._dependency_groups.keys()),
                 optional=self.optional,
                 base=self.root_dir,
@@ -493,28 +526,32 @@ class Package(PackageSpecification):
                 extras=self.features,
             )
         elif self.source_type == "file":
+            assert self._source_url is not None
             dep = FileDependency(
                 self._name,
-                Path(cast(str, self._source_url)),
+                Path(self._source_url),
+                directory=self.source_subdirectory,
                 groups=list(self._dependency_groups.keys()),
                 optional=self.optional,
                 base=self.root_dir,
                 extras=self.features,
             )
         elif self.source_type == "url":
+            assert self._source_url is not None
             dep = URLDependency(
                 self._name,
-                cast(str, self._source_url),
+                self._source_url,
                 directory=self.source_subdirectory,
                 groups=list(self._dependency_groups.keys()),
                 optional=self.optional,
                 extras=self.features,
             )
         elif self.source_type == "git":
+            assert self._source_url is not None
             dep = VCSDependency(
                 self._name,
                 self.source_type,
-                cast(str, self.source_url),
+                self._source_url,
                 rev=self.source_reference,
                 resolved_rev=self.source_resolved_reference,
                 directory=self.source_subdirectory,
@@ -537,32 +574,53 @@ class Package(PackageSpecification):
 
         return dep.with_constraint(self._version)
 
-    @contextmanager
-    def with_python_versions(self, python_versions: str) -> Iterator[None]:
-        original_python_versions = self.python_versions
-
-        self.python_versions = python_versions
-
-        yield
-
-        self.python_versions = original_python_versions
-
     def satisfies(
         self, dependency: Dependency, ignore_source_type: bool = False
     ) -> bool:
         """
         Helper method to check if this package satisfies a given dependency.
 
-        This is determined by assessing if this instance provides the package and
-        features specified by the given dependency. Further, version and source
-        types are checked.
+        This is determined by assessing if this instance provides the package specified
+        by the given dependency. Further, version and source types are checked.
         """
-        if not self.provides(dependency) or not dependency.constraint.allows(
-            self.version
-        ):
+        if self.name != dependency.name:
             return False
 
-        return ignore_source_type or self.is_same_source_as(dependency)
+        if not dependency.constraint.allows(self.version):
+            return False
+
+        if not (ignore_source_type or self.source_satisfies(dependency)):  # noqa: SIM103
+            return False
+
+        return True
+
+    def source_satisfies(self, dependency: Dependency) -> bool:
+        """Determine whether this package's source satisfies the given dependency."""
+        if dependency.source_type is None:
+            if dependency.source_name is None:
+                # The dependency doesn't care about the source, so this package
+                # certainly satisfies it.
+                return True
+
+            # The dependency specifies a source_name but not a type: it wants either
+            # pypi or a legacy repository.
+            #
+            # - If this package has no source type then it's from pypi, so it
+            #   matches if and only if that's what the dependency wants
+            # - Else this package is a match if and only if it is from the desired
+            #   repository
+            if self.source_type is None:
+                return dependency.source_name.lower() == "pypi"
+
+            return (
+                self.source_type == "legacy"
+                and self.source_reference is not None
+                and self.source_reference.lower() == dependency.source_name.lower()
+            )
+
+        # The dependency specifies a source: this package matches if and only if it is
+        # from that source.
+        return dependency.is_same_source_as(self)
 
     def __eq__(self, other: object) -> bool:
         if not isinstance(other, Package):
@@ -577,24 +635,30 @@ class Package(PackageSpecification):
         return f"{self.complete_name} ({self.full_pretty_version})"
 
     def __repr__(self) -> str:
-        args = [repr(self._name), repr(self._version.text)]
+        args = [repr(self._pretty_name), repr(self._version.text)]
 
         if self._features:
-            args.append(f"features={repr(self._features)}")
+            args.append(f"features={self._features!r}")
 
         if self._source_type:
-            args.append(f"source_type={repr(self._source_type)}")
-            args.append(f"source_url={repr(self._source_url)}")
+            args.append(f"source_type={self._source_type!r}")
+            args.append(f"source_url={self._source_url!r}")
 
             if self._source_reference:
-                args.append(f"source_reference={repr(self._source_reference)}")
+                args.append(f"source_reference={self._source_reference!r}")
 
             if self._source_resolved_reference:
                 args.append(
-                    f"source_resolved_reference={repr(self._source_resolved_reference)}"
+                    f"source_resolved_reference={self._source_resolved_reference!r}"
                 )
             if self._source_subdirectory:
-                args.append(f"source_subdirectory={repr(self._source_subdirectory)}")
+                args.append(f"source_subdirectory={self._source_subdirectory!r}")
+
+        if self.develop:
+            args.append(f"develop={self.develop}")
+
+        if self._yanked:
+            args.append(f"yanked={self.yanked}")
 
         args_str = ", ".join(args)
         return f"Package({args_str})"
